// -*- C++ -*-
//
//  This file is part of WEEVILGEN.
//
//  Authors: See the file AUTHORS for full details.
//
//  Copyright (C) 2002-2004 University of Colorado
//
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307,
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: random_browser.cc,v 1.1 2007-07-10 13:08:07 wolf Exp $
//
// implementation of a random, communicating browser
#include <cassert>
#include <cstring>
#include <string>
#include <weevil/weevil.h>

using namespace ssim;
using namespace std;
using namespace weevil;

/**
 * application-specific function and classes
 */
unsigned int get_rand( unsigned int a, unsigned int b )
{
	// We use the following formula to scale rand() to
	// the range of values we are looking for a -> b inclusive:
	// x = a + int( (rand() / (RAND_MAX+1.0))*(b-a+1) )
	return( a + (unsigned int)(((double)rand()/(RAND_MAX+1.0))*(double)(b-a+1)) );
}

class neighborhood{
public:
  std::string id;
  std::string eventtype;
  int signaldelay;
  double probability;
};

/**
 * Event classes
 */
class Timeout : public Event {};

class Recommendation : public Event
{
public:
  std::string from;
  std::string type;
  unsigned int fileid;

  Recommendation( std::string f, std::string t, unsigned int fi ) : from ( f ), type( t ), fileid( fi ) {}
};

/**
 * reactive process class
 */
class RandomBrowser : public weevil::WeevilProcess
{
  /**
   * inputted properties from m4 configuration
   */
  unsigned int	        m_num_files;
  unsigned int		m_num_gets;
  Time			m_get_timeout;
  Time                  m_delay;
  unsigned int          m_file_start;
  unsigned int          m_file_end;

  /**
   * data structure to be used
   */
  unsigned int 		m_left_gets;
  vector<neighborhood*> m_get_neighbors;

public:
  /**
   * reactive process, thus no implementation for "main" method
   */
  RandomBrowser();

  virtual void main(){};

  virtual void init();

  virtual void process_event( const Event* e );

  /**
   * assignment of process properties
   */
  void num_files( unsigned int n )
  { m_num_files = n; }

  void num_gets( unsigned int n )
  { m_num_gets = n; m_left_gets = n; }

  void get_timeout( Time t )
  { m_get_timeout = t; }

  void delay( Time t )
  { m_delay = t; }

  void file_range( std::string range ){
    int begin = 0;
    int end = range.find_first_of("-", begin);
    m_file_start = atoi(range.substr(begin, end-begin).c_str());
    begin = end + 1;
    m_file_end = atoi(range.substr(begin).c_str());
  }

  neighborhood* getobject( std::string neighborstr ){
    int begin = 0;
    int end = neighborstr.find_first_of(" ", begin);
    neighborhood* a_neighbor = new neighborhood();
    a_neighbor->id = neighborstr.substr(begin, end-begin);
    begin = end + 1;
    end = neighborstr.find_first_of(" ", begin);
    a_neighbor->eventtype = neighborstr.substr(begin, end-begin);
    begin = end + 1;
    end = neighborstr.find_first_of(" ", begin);
    a_neighbor->signaldelay = atoi(neighborstr.substr(begin, end-begin).c_str());
    begin = end + 1;
    a_neighbor->probability = atof(neighborstr.substr(begin).c_str());
    return a_neighbor;
  }

  void get_neighbors( std::string list ){
    if ( list.compare("") != 0 ){
      neighborhood* one_neighbor = new neighborhood();
      int begin = 0;
      int end = list.find_first_of("|", begin);
      std::string one_neighborstring;
      while (end != string::npos){
	one_neighborstring = list.substr(begin, end-begin);
	begin = list.find_first_not_of(" ", end + 1);
	end = list.find_first_of("|", begin);
	one_neighbor = getobject(one_neighborstring);
	m_get_neighbors.push_back(one_neighbor);
      }
    }
  }
};

/**
 * method implementation
 */
RandomBrowser :: RandomBrowser(){
  m_delay = 0;
  m_get_neighbors.clear();
  m_file_start = 0;
}

void RandomBrowser :: init()
{
  assert( m_num_files && m_get_timeout > 0.0 && m_delay >= 0.0);

  // set the initial timeout
  if ( m_left_gets > 0 ){
  	m_left_gets--;
  	WeevilGen::self_signal_event( new Timeout(), m_delay + m_get_timeout );
  }
}

void RandomBrowser :: process_event( const Event* e )
{
  const Timeout * t;
  const Recommendation * r;

  if ((t = dynamic_cast<const Timeout *>(e)) != 0) {
    // a timeout event, so just pick a random number <= n and issue an output
    unsigned int fid = get_rand( 1, m_num_files )+m_file_start-1;

    char msg[256];
    sprintf( msg, "GET(file%d)", fid );
    workload_output( msg );

    // randomly tell neighbors about the file
    for (int fi = 0; fi < m_get_neighbors.size(); fi ++){
      if ((drand48() < m_get_neighbors[fi]->probability)&&(m_get_neighbors[fi]->id != id())){
	WeevilGen::signal_event( m_get_neighbors[fi]->id, new Recommendation( id(), "GET", fid ), m_get_neighbors[fi]->signaldelay );
      }
    }
    WeevilGen::self_signal_event( new Timeout(), m_get_timeout );
  }
  else if ((r = dynamic_cast<const Recommendation *>(e)) != 0){
    // a recommendation event from someone else
    const Recommendation* rec = static_cast<const Recommendation*>( e );
    std::string from = rec->from;

    char msg[256];    
    sprintf( msg, "GET(file%d)", rec->fileid );
    workload_output( msg );
  }
}
