// -*- C++ -*-
//
//  This file is part of WEEVIL.
//
//  Authors: See the file AUTHORS for full details.
//
//  Copyright (C) 2002-2006 University of Colorado
//
//  Weevil is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  Weevil is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
//  USA, or send email to serl@cs.colorado.edu.
//
// $Id: weevil.h,v 1.1 2007-07-10 13:08:09 wolf Exp $
//
#ifndef WEEVIL_WEEVIL_H
#define WEEVIL_WEEVIL_H

#include <iostream>
#include <map>
#include <set>
#include <siena/ssim.h>
#include <siena/tprocess.h>
#include <string>

/** Provides services needed by programs that comprise a weevil
 * workload generator.
 */
namespace weevil
{
  /** @brief An extension of <code>ssim::Sim</code> that has some
   * convenience methods for dealing with processes by ids.
   */
  class WeevilGen : public ssim::Sim
  {
    static std::map<std::string,ssim::ProcessId> m_id2pid;
    static std::map<ssim::ProcessId,std::string> m_pid2id;    

  public:
    // Adds process to the internal maps of processes.
    //
    static void add_process( const std::string& id, ssim::ProcessId pid )
    {
      m_id2pid.insert( make_pair( id, pid ) );
      m_pid2id.insert( make_pair( pid, id ) );
    }

    /** @brief Broadcasts event to all processes.
     */
    static void broadcast_event( const ssim::Event* event, ssim::Time lag = 0 );

    // Static method to return the ProcessId given a
    // particular identifier.
    //
    static ssim::ProcessId id2pid( const std::string& id );

    // returns the id of the process that has
    // number pid.
    //
    static const std::string& pid2id( ssim::ProcessId pid );

    /** @brief Signals event to itself
     */
    static void self_signal_event( const ssim::Event* event, ssim::Time lag = 0 );

    /** @brief Signals event to neighbor based on its ID
     */
    static void signal_event( const std::string& id, const ssim::Event* event, ssim::Time lag = 0 );
  };

  /** @brief A class providing some convenience methods for workload
   *  generation (such as, dealing with processes by ids, workload
   *  output).
   */
  class WeevilActor
  {
    std::string		m_id;
    ssim::ProcessId	m_pid;

  public:
    // Sets the id of the process.
    //
    void id( const std::string& id )
    { m_id = id; }

    // Returns the id of the process.
    //
    const std::string& id() const
    { return m_id; }

    // Sets the processes processid.
    //
    void pid( ssim::ProcessId pid )
    { m_pid = pid; }

    // Returns the pid of the process.
    //
    ssim::ProcessId pid() const
    { return m_pid; }

    /** @brief Outputs a workload event.
     */
    void workload_output( const char* msg, const char* condition, const char* variable ) const
    { std::cout << "event(" << ssim::Sim::clock() << "," << m_id << "," << msg << "," << condition << "," << variable << ")" << std::endl; }
  };

  /** @brief An extension of <code>ssim::Process</code> and
   *  <code>weevil::WeevilActor</code>.
   */
  class WeevilProcess : public ssim::Process, public weevil::WeevilActor
  {
  public:
    // Constructor chains to superclass.
    //
    WeevilProcess() : Process() {}
  };

  /** @brief An extension of <code>ssim::TProcess</code> that and
   * <code>weevil::WeevilActor</code>.
   */
  class WeevilTProcess : public ssim::TProcess, public weevil::WeevilActor
  {
  public:
    // Constructor chains to superclass.
    //
    WeevilTProcess() : TProcess() {}
  
    /** @brief Accepts an event over a given interval.
     */
    static const ssim::Event * wait_for_event(ssim::Time timeout = ssim::INIT_TIME)
    {
      return ssim::TProcess::wait_for_event(timeout);
    }
  };

  /** @brief Simple simulation process that kills the simulation after
   * a specified number of virtual time units.
   */
  class KillerProcess : public ssim::Process
    {
    private:
      unsigned int m_delay;
    public:
      KillerProcess( unsigned int delay )
	: m_delay( delay ) {}
      
      void init()
      {
	ssim::Sim::set_stop_time( m_delay );
      }
  };
};

#endif
